@@ -73,7 +73,7 @@ struct Proxy {
   GUPnPServiceProxy *proxy;
 
   gchar *external_ip;
-  GUPnPServiceProxyAction *external_ip_action;
+  GCancellable *external_ip_cancellable;
   gboolean external_ip_failed;
 
   GPtrArray *proxymappings;
@@ -92,7 +92,7 @@ struct ProxyMapping {
   struct Proxy *proxy;
   struct Mapping *mapping;
 
-  GUPnPServiceProxyAction *action;
+  GCancellable *cancellable;
 
   gboolean mapped;
   guint actual_external_port;
@@ -124,7 +124,8 @@ guint signals[LAST_SIGNAL] = { 0 };
    GUPnPSimpleIgdPrivate))
 
 
-G_DEFINE_TYPE (GUPnPSimpleIgd, gupnp_simple_igd, G_TYPE_OBJECT);
+G_DEFINE_TYPE_WITH_CODE (GUPnPSimpleIgd, gupnp_simple_igd, G_TYPE_OBJECT,
+    G_ADD_PRIVATE (GUPnPSimpleIgd));
 
 
 static void gupnp_simple_igd_constructed (GObject *object);
@@ -171,8 +172,6 @@ gupnp_simple_igd_class_init (GUPnPSimpleIgdClass *klass)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
-  g_type_class_add_private (klass, sizeof (GUPnPSimpleIgdPrivate));
-
   gobject_class->constructed = gupnp_simple_igd_constructed;
   gobject_class->dispose = gupnp_simple_igd_dispose;
   gobject_class->finalize = gupnp_simple_igd_finalize;
@@ -264,7 +263,7 @@ gupnp_simple_igd_class_init (GUPnPSimpleIgdClass *klass)
 static void
 gupnp_simple_igd_init (GUPnPSimpleIgd *self)
 {
-  self->priv = GUPNP_SIMPLE_IGD_GET_PRIVATE (self);
+  self->priv = gupnp_simple_igd_get_instance_private (self);
 
   self->priv->service_proxies = g_ptr_array_new ();
   self->priv->mappings = g_ptr_array_new ();
@@ -365,27 +364,28 @@ _external_ip_address_changed (GUPnPServiceProxy *proxy, const gchar *variable,
 }
 
 static void
-_service_proxy_delete_port_mapping (GUPnPServiceProxy *proxy,
-    GUPnPServiceProxyAction *action,
+_service_proxy_delete_port_mapping (GObject *source_object, GAsyncResult *res,
     gpointer user_data)
 {
+  GUPnPServiceProxy *proxy = GUPNP_SERVICE_PROXY (source_object);
+  GUPnPServiceProxyAction *action;
   GError *error = NULL;
   GUPnPSimpleIgd *self = user_data;
 
+  action = gupnp_service_proxy_call_action_finish (proxy, res, &error);
 
-  if (!gupnp_service_proxy_end_action (proxy, action, &error,
-          NULL))
-  {
+  if (action == NULL ||
+      !gupnp_service_proxy_action_get_result (action, &error, NULL)) {
     g_return_if_fail (error);
     g_warning ("Error deleting port mapping: %s", error->message);
   }
   g_clear_error (&error);
 
-  if (self)
-  {
-    self->priv->deleting_count--;
-    g_object_unref (self);
-  }
+  if (action)
+    gupnp_service_proxy_action_unref (action);
+
+  self->priv->deleting_count--;
+  g_object_unref (self);
 }
 
 static void
@@ -395,15 +395,19 @@ free_proxymapping (struct ProxyMapping *pm, GUPnPSimpleIgd *self)
 
   if (pm->mapped && self)
   {
+    GUPnPServiceProxyAction *action;
+
     self->priv->deleting_count++;
     g_object_ref (self);
-    gupnp_service_proxy_begin_action (pm->proxy->proxy,
-        "DeletePortMapping",
-        _service_proxy_delete_port_mapping, self,
+
+    action = gupnp_service_proxy_action_new ("DeletePortMapping",
         "NewRemoteHost", G_TYPE_STRING, "",
         "NewExternalPort", G_TYPE_UINT, pm->actual_external_port,
         "NewProtocol", G_TYPE_STRING, pm->mapping->protocol,
         NULL);
+
+    gupnp_service_proxy_call_action_async (pm->proxy->proxy, action, NULL,
+        _service_proxy_delete_port_mapping, self);
   }
 
   g_slice_free (struct ProxyMapping, pm);
@@ -412,8 +416,8 @@ free_proxymapping (struct ProxyMapping *pm, GUPnPSimpleIgd *self)
 static void
 free_proxy (struct Proxy *prox)
 {
-  if (prox->external_ip_action)
-    gupnp_service_proxy_cancel_action (prox->proxy, prox->external_ip_action);
+  g_cancellable_cancel (prox->external_ip_cancellable);
+  g_clear_object (&prox->external_ip_cancellable);
 
   gupnp_service_proxy_remove_notify (prox->proxy, "ExternalIPAddress",
       _external_ip_address_changed, prox);
@@ -582,7 +586,6 @@ static void
 gupnp_simple_igd_constructed (GObject *object)
 {
   GUPnPSimpleIgd *self = GUPNP_SIMPLE_IGD_CAST (object);
-  SoupSession *session;
 
   self->priv->main_context = g_main_context_get_thread_default ();
   if (!self->priv->main_context)
@@ -615,65 +618,77 @@ gupnp_simple_igd_new (void)
 
 
 static void
-_service_proxy_got_external_ip_address (GUPnPServiceProxy *proxy,
-    GUPnPServiceProxyAction *action,
-    gpointer user_data)
+_service_proxy_got_external_ip_address (GObject *source_object,
+    GAsyncResult *res, gpointer user_data)
 {
+  GUPnPServiceProxy *proxy = GUPNP_SERVICE_PROXY (source_object);
   struct Proxy *prox = user_data;
   GUPnPSimpleIgd *self = prox->parent;
+  GUPnPServiceProxyAction *action;
   GError *error = NULL;
   gchar *ip = NULL;
+  guint i;
 
-  g_return_if_fail (prox->external_ip_action == action);
+  action = gupnp_service_proxy_call_action_finish (proxy, res, &error);
 
-  prox->external_ip_action = NULL;
+  if (action == NULL &&
+      g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+    return;
 
-  if (gupnp_service_proxy_end_action (proxy, action, &error,
-          "NewExternalIPAddress", G_TYPE_STRING, &ip,
-          NULL))
-  {
-    guint i;
+  g_clear_object (&prox->external_ip_cancellable);
 
-    if (!validate_ip_address (ip))
-    {
-      prox->external_ip_failed = TRUE;
+  if (action == NULL)
+    goto error;
 
-      for (i=0; i < prox->proxymappings->len; i++)
-      {
-        struct ProxyMapping *pm = g_ptr_array_index (prox->proxymappings, i);
-        GError gerror = {GUPNP_SIMPLE_IGD_ERROR,
-                         GUPNP_SIMPLE_IGD_ERROR_EXTERNAL_ADDRESS,
-                         "Invalid IP address returned by router"};
-
-        g_signal_emit (self, signals[SIGNAL_ERROR_MAPPING_PORT],
-            GUPNP_SIMPLE_IGD_ERROR, &gerror, pm->mapping->protocol,
-            pm->mapping->requested_external_port, pm->mapping->local_ip,
-            pm->mapping->local_port, pm->mapping->description);
-      }
-      return;
-    }
-
-    /* Only emit the new signal if the IP changes */
-    if (prox->external_ip &&
-        strcmp (ip, prox->external_ip))
-    {
-      for (i=0; i < prox->proxymappings->len; i++)
-      {
-        struct ProxyMapping *pm = g_ptr_array_index (prox->proxymappings, i);
-
-        if (pm->mapped)
-          g_signal_emit (self, signals[SIGNAL_MAPPED_EXTERNAL_PORT], 0,
-              pm->mapping->protocol, ip, prox->external_ip,
-              pm->actual_external_port, pm->mapping->local_ip,
-              pm->mapping->local_port, pm->mapping->description);
-      }
-    }
-
-    g_free (prox->external_ip);
-    prox->external_ip = ip;
+  if (!gupnp_service_proxy_action_get_result (action, &error,
+          "NewExternalIPAddress", G_TYPE_STRING, &ip, NULL)) {
+    gupnp_service_proxy_action_unref (action);
+    goto error;
   }
-  else
+  gupnp_service_proxy_action_unref (action);
+
+  if (!validate_ip_address (ip))
   {
+    prox->external_ip_failed = TRUE;
+
+    for (i=0; i < prox->proxymappings->len; i++)
+    {
+      struct ProxyMapping *pm = g_ptr_array_index (prox->proxymappings, i);
+      GError gerror = {GUPNP_SIMPLE_IGD_ERROR,
+                       GUPNP_SIMPLE_IGD_ERROR_EXTERNAL_ADDRESS,
+                       "Invalid IP address returned by router"};
+
+      g_signal_emit (self, signals[SIGNAL_ERROR_MAPPING_PORT],
+          GUPNP_SIMPLE_IGD_ERROR, &gerror, pm->mapping->protocol,
+          pm->mapping->requested_external_port, pm->mapping->local_ip,
+          pm->mapping->local_port, pm->mapping->description);
+    }
+    return;
+  }
+
+  /* Only emit the new signal if the IP changes */
+  if (prox->external_ip &&
+      strcmp (ip, prox->external_ip))
+  {
+    for (i=0; i < prox->proxymappings->len; i++)
+    {
+      struct ProxyMapping *pm = g_ptr_array_index (prox->proxymappings, i);
+
+      if (pm->mapped)
+        g_signal_emit (self, signals[SIGNAL_MAPPED_EXTERNAL_PORT], 0,
+            pm->mapping->protocol, ip, prox->external_ip,
+            pm->actual_external_port, pm->mapping->local_ip,
+            pm->mapping->local_port, pm->mapping->description);
+    }
+  }
+
+  g_free (prox->external_ip);
+  prox->external_ip = ip;
+
+  return;
+
+error:
+{
     guint i;
 
     prox->external_ip_failed = TRUE;
@@ -696,9 +711,17 @@ static void
 gupnp_simple_igd_gather (GUPnPSimpleIgd *self,
     struct Proxy *prox)
 {
-  prox->external_ip_action = gupnp_service_proxy_begin_action (prox->proxy,
-      "GetExternalIPAddress",
-      _service_proxy_got_external_ip_address, prox, NULL);
+  GUPnPServiceProxyAction *action;
+
+  g_assert (prox->external_ip_cancellable == NULL);
+  prox->external_ip_cancellable = g_cancellable_new ();
+
+  action = gupnp_service_proxy_action_new (
+      "GetExternalIPAddress", NULL);
+
+  gupnp_service_proxy_call_action_async (prox->proxy, action,
+      prox->external_ip_cancellable,
+      _service_proxy_got_external_ip_address, prox);
 
   gupnp_service_proxy_add_notify (prox->proxy, "ExternalIPAddress",
       G_TYPE_STRING, _external_ip_address_changed, prox);
@@ -707,42 +730,52 @@ gupnp_simple_igd_gather (GUPnPSimpleIgd *self,
 }
 
 static void
-_service_proxy_renewed_port_mapping (GUPnPServiceProxy *proxy,
-    GUPnPServiceProxyAction *action,
+_service_proxy_renewed_port_mapping (GObject *source_object, GAsyncResult *res,
     gpointer user_data)
 {
+  GUPnPServiceProxy *proxy = GUPNP_SERVICE_PROXY (source_object);
+  GUPnPServiceProxyAction *action;
   struct ProxyMapping *pm = user_data;
   GUPnPSimpleIgd *self = pm->proxy->parent;
   GError *error = NULL;
 
-  g_return_if_fail (pm->action == action);
+  action = gupnp_service_proxy_call_action_finish (proxy, res, &error);
 
-  pm->action = NULL;
+  if (action == NULL &&
+      g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+    return;
 
-  if (!gupnp_service_proxy_end_action (proxy, action, &error,
-          NULL))
-  {
-    g_return_if_fail (error);
-    g_signal_emit (self, signals[SIGNAL_ERROR_MAPPING_PORT], error->domain,
-        error, pm->mapping->protocol, pm->mapping->requested_external_port,
-        pm->mapping->local_ip, pm->mapping->local_port,
-        pm->mapping->description);
+  g_clear_object (&pm->cancellable);
+
+  if (action) {
+    if (gupnp_service_proxy_action_get_result (action, &error, NULL)) {
+      gupnp_service_proxy_action_unref (action);
+      return;
+    }
+    gupnp_service_proxy_action_unref (action);
   }
+
+  g_return_if_fail (error);
+  g_signal_emit (self, signals[SIGNAL_ERROR_MAPPING_PORT], error->domain,
+      error, pm->mapping->protocol, pm->mapping->requested_external_port,
+      pm->mapping->local_ip, pm->mapping->local_port,
+      pm->mapping->description);
   g_clear_error (&error);
 }
 
 static void
 gupnp_simple_igd_call_add_port_mapping (struct ProxyMapping *pm,
-    GUPnPServiceProxyActionCallback callback)
+    GAsyncReadyCallback callback)
 {
+  GUPnPServiceProxyAction *action;
   g_assert (pm);
-  g_return_if_fail (pm->action == NULL);
+  g_return_if_fail (pm->cancellable == NULL);
   g_assert (pm->proxy);
   g_assert (pm->mapping);
 
-  pm->action = gupnp_service_proxy_begin_action (pm->proxy->proxy,
-      "AddPortMapping",
-      callback, pm,
+  pm->cancellable = g_cancellable_new ();
+
+  action = gupnp_service_proxy_action_new ("AddPortMapping",
       "NewRemoteHost", G_TYPE_STRING, "",
       "NewExternalPort", G_TYPE_UINT, pm->actual_external_port,
       "NewProtocol", G_TYPE_STRING, pm->mapping->protocol,
@@ -752,6 +785,9 @@ gupnp_simple_igd_call_add_port_mapping (struct ProxyMapping *pm,
       "NewPortMappingDescription", G_TYPE_STRING, pm->mapping->description,
       "NewLeaseDuration", G_TYPE_UINT, pm->mapping->lease_duration,
       NULL);
+
+  gupnp_service_proxy_call_action_async (pm->proxy->proxy, action,
+      pm->cancellable, callback, pm);
 }
 
 static gboolean
@@ -768,39 +804,60 @@ _renew_mapping_timeout (gpointer user_data)
 }
 
 static void
-_service_proxy_added_port_mapping (GUPnPServiceProxy *proxy,
-    GUPnPServiceProxyAction *action,
+_service_proxy_added_port_mapping (GObject *source_object, GAsyncResult *res,
     gpointer user_data)
 {
+  GUPnPServiceProxy *proxy = GUPNP_SERVICE_PROXY (source_object);
+  GUPnPServiceProxyAction *action;
   struct ProxyMapping *pm = user_data;
-  GUPnPSimpleIgd *self = pm->proxy->parent;
+  GUPnPSimpleIgd *self;
   GError *error = NULL;
 
-  g_return_if_fail (pm->action == action);
+  /* This is a hack, but wek now that "res" is actually implemented as a GTask,
+   * we're just too lazy to carry our own reference counted structure
+   */
+  if (g_cancellable_is_cancelled (g_task_get_cancellable (G_TASK (res))))
+    return;
 
-  pm->action = NULL;
+  action = gupnp_service_proxy_call_action_finish (proxy, res, &error);
 
-  if (gupnp_service_proxy_end_action (proxy, action, &error,
-          NULL))
-  {
-    pm->mapped = TRUE;
+  if (action == NULL &&
+      g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+    return;
 
-    if (pm->proxy->external_ip)
-      g_signal_emit (self, signals[SIGNAL_MAPPED_EXTERNAL_PORT], 0,
-          pm->mapping->protocol, pm->proxy->external_ip, NULL,
-          pm->actual_external_port, pm->mapping->local_ip,
-          pm->mapping->local_port, pm->mapping->description);
+  self = pm->proxy->parent;
+  g_clear_object (&pm->cancellable);
 
-    if (pm->mapping->lease_duration > 0)
-    {
-      pm->renew_src =
-        g_timeout_source_new_seconds (pm->mapping->lease_duration / 2);
-      g_source_set_callback (pm->renew_src,
-          _renew_mapping_timeout, pm, NULL);
-      g_source_attach (pm->renew_src, self->priv->main_context);
-    }
+  if (action == NULL)
+    goto error;
+
+  if (!gupnp_service_proxy_action_get_result (action, &error, NULL)) {
+    gupnp_service_proxy_action_unref (action);
+    goto error;
   }
-  else
+
+  gupnp_service_proxy_action_unref (action);
+
+  pm->mapped = TRUE;
+
+  if (pm->proxy->external_ip)
+    g_signal_emit (self, signals[SIGNAL_MAPPED_EXTERNAL_PORT], 0,
+        pm->mapping->protocol, pm->proxy->external_ip, NULL,
+        pm->actual_external_port, pm->mapping->local_ip,
+        pm->mapping->local_port, pm->mapping->description);
+
+  if (pm->mapping->lease_duration > 0)
+  {
+    pm->renew_src =
+      g_timeout_source_new_seconds (pm->mapping->lease_duration / 2);
+    g_source_set_callback (pm->renew_src,
+        _renew_mapping_timeout, pm, NULL);
+    g_source_attach (pm->renew_src, self->priv->main_context);
+  }
+
+  return;
+
+ error:
   {
     g_return_if_fail (error);
 
@@ -1052,10 +1109,8 @@ gupnp_simple_igd_remove_port_local (GUPnPSimpleIgd *self,
 static void
 stop_proxymapping (struct ProxyMapping *pm, gboolean stop_renew)
 {
-  if (pm->action)
-    gupnp_service_proxy_cancel_action (pm->proxy->proxy,
-        pm->action);
-  pm->action = NULL;
+  g_cancellable_cancel (pm->cancellable);
+  g_clear_object (&pm->cancellable);
 
   if (stop_renew && pm->renew_src)
   {
